﻿/*
 * This file is part of the ALT project, module Alt-Customize-Branding.
 * Copyright (c) BaseALT Ltd 2020 Pavel Moseev <mars@altlinux.org>
 *
 * This is free software; you can use this application under the GNU
 * General Public License, version 3. See the file "COPYING" for the
 * exact licensing terms.
 */

#include "mainwindow.h"
#include "./ui_mainwindow.h"
#include <QtWidgets>
#include <QPixmap>
#include <QLabel>
#include <QErrorMessage>
#include <QMessageBox>
#include <QFile>
#include <QDir>
#include <QSettings>
#include <QVariant>
#include <QMap>
#include <QTextStream>
#include <QFileInfo>

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
    , ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    QDesktopWidget desktop; // Set the main window in the desktop center
    QRect rect = desktop.availableGeometry(this);
    QPoint center = rect.center();
    int x = center.x() - (width()/2);
    int y = center.y() - (height()/2);
    center.setX(x);
    center.setY(y);
    move(center);

    scene = new QGraphicsScene(this);
    myTimerId_ = 0;

    qtranslator = 0;
    app_translator = 0;
    useTranslator(qtranslator, QLatin1String("qtbase"));
    useTranslator(app_translator, QLatin1String("alt-customize-branding"));
    ui->retranslateUi(this);

    connect(ui->button_Close, SIGNAL (clicked()), this, SLOT (closeApplication()));
    connect(ui->button_Apply, SIGNAL (clicked()), this, SLOT (runApply()));
    connect(ui->button_viewPlymouthImage, SIGNAL (clicked()), this, SLOT (viewPlymouthImageF()));
    connect(ui->button_openPlymouthImage, SIGNAL (clicked()), this, SLOT (openPlymouthImageF()));
    connect(ui->button_viewNewLogo, SIGNAL (clicked()), this, SLOT (viewNewLogoF()));
    connect(ui->button_openNewLogo, SIGNAL (clicked()), this, SLOT (openNewLogoF()));
    connect(ui->button_Clear, SIGNAL (clicked()), this, SLOT (cleanupForm()));
}

MainWindow::~MainWindow()
{
    delete ui;
}

QString MainWindow::openFileToChangeBranding(QString pathFromLineEdit) // OPEN FILE
{
    QString currentFolder;
    if (!pathFromLineEdit.isEmpty())
    {
        currentFolder = pathFromLineEdit; // Open last opening directory
    }
    else currentFolder = ""; // Open current directory with program
    QString fileTCBPath = QFileDialog::getOpenFileName(
                this,
                tr("Open File"),
                currentFolder,
                tr("All (*);;JPEG (*.jpg *.jpeg);;PNG (*.png)")
                );
    if (fileTCBPath.isEmpty())
    {
        fileTCBPath = currentFolder;
    }
    return (fileTCBPath);
}

void MainWindow::resizeEvent(QResizeEvent *event) // Resize main window => resize image and logo.
{
    Q_UNUSED(event)
    if (!fileTCBPathInScene_.isEmpty()) // Scaled image size
    {
        ui->graphicsView->scene()->clear();
        viewFileToChangeBranding(fileTCBPathInScene_);
        if (logoInScene_) // Logo test
        {
            viewLogoToChangeBranding();
        }
    }
}

void MainWindow::viewFileToChangeBranding(QString fileTCBPath) // VIEW BACKGROUND
{
    if (!fileTCBPath.isEmpty())
    {
        fileTCBPathInScene_ = QLatin1String("");
        imageObject = new QImage();
        imageObject->load(fileTCBPath);
        QPixmap pixmap(fileTCBPath);
        image = QPixmap::fromImage(*imageObject);
        scene->addPixmap(image); // QGraphicsPixmapItem* pPixmapItem2 = scene->addPixmap(image);
        scene->setSceneRect(image.rect());
        ui->graphicsView->setScene(scene);
        fileTCBPathInScene_ = fileTCBPath;
        setScene_ = true;
        addBackgroundBtnActivated_ = true;
    }
}

void MainWindow::viewLogoToChangeBranding() // VIEW LOGO
{
    QString logoPath = ui->lineEdit_newLogo->text();
    if (!logoPath.isEmpty())
    {
        QImage logoImage(logoPath);
        item = new QGraphicsPixmapItem(QPixmap::fromImage(logoImage));
        item->setZValue(100);
        item->setAcceptTouchEvents(true); // item->setAcceptHoverEvents(true);
        item->setFlags(QGraphicsItem::ItemIsSelectable|QGraphicsItem::ItemIsMovable);

        if (addLogoButtonActivated_) // Repaint logo on a new backgroung image
        {
            item->setPos(currentItemX_, currentItemY_);
        }
        scene->addItem(item);
        ui->graphicsView->setScene(scene);
        logoInScene_ = true; // isItemAdded
        setScene_ = true;
    }
    else logoInScene_ = false;
}

void MainWindow::runApply() // Apply button
{
    ui->label_wait->setText(QLatin1String(""));
    QString newImage;
    QString imageFile = ui->lineEdit_plymouth->text();
    QString logoFile  = ui->lineEdit_newLogo->text();
    if (imageFile.isEmpty() || logoFile.isEmpty())
    {
        QMessageBox::information(this, tr("Warning"), tr("Logo or Image not set!"));
        newImage = QLatin1String("not_set");
    }
    else
    {
        // Create work directory:
        QString tempDir = appTmpSettingsDir_;
        QDir tDir(tempDir);
        if (!tDir.exists())
        {
            QDir().mkdir(tempDir);
        }
        newImage = createNewImage(imageFile, logoFile);
        QString resizedImage = resizeImage(newImage, 800, 600);   // Resize Grub Image to 800x600.
        QString convertedImage = convertImageToJpg(resizedImage); // Convert png grub image to jpg.

        // Parse image file name:
        QFileInfo fileInfoImg(imageFile);
        QString imageFileName(fileInfoImg.fileName()); // Background image name
        QFileInfo fileInfoLogo(logoFile);
        QString logoFileName(fileInfoLogo.fileName()); // Logo image name

        // Copy source images to appTmpSettingsDir_:
        QFile::copy(imageFile, appTmpSettingsDir_ + imageFileName);
        QFile::copy(logoFile, appTmpSettingsDir_ + logoFileName);

        // If file exists, then use. Else create new .ini-file:
        QString acbConfigFile = appTmpSettingsDir_ + QStringLiteral("alt-customize-branding-settings.ini");
        QFile confFile(acbConfigFile);
        if (!confFile.exists()) // Create .ini-file
        {
            QSettings acbSettings(acbConfigFile, QSettings::IniFormat);
            acbSettings.setValue("Main/AppName", appName_);                 // "alt-customize-branding"
            acbSettings.setValue("Main/TempConfigDir", appTmpSettingsDir_); // "/tmp/alt-customize-branding/"
            acbSettings.setValue("Main/ConfigDir", appSettingsDir_);        // "/var/lib/alt-customize-branding/"
            acbSettings.setValue("Source/LogoImage",logoFileName);
            acbSettings.setValue("Source/BackgroundImage",imageFileName);
            acbSettings.setValue("Modified/BigPngImage","newimage.png");
            acbSettings.setValue("Modified/ResizedPngImage","newimage800x600.png");
            acbSettings.setValue("Modified/ResizedJpgImage","newimage800x600.jpg");
            acbSettings.setValue("Modified/LogoCoordsX",currentItemX_);
            acbSettings.setValue("Modified/LogoCoordsY",currentItemY_);
            acbSettings.setValue("PreviousTheme/ThemeName","");
        }
        else // Fill .ini-file with values
        {
            QSettings acbSettings(acbConfigFile, QSettings::IniFormat);
            acbSettings.beginGroup("Source");
            acbSettings.setValue("LogoImage", logoFileName);
            acbSettings.setValue("BackgroundImage", imageFileName);
            acbSettings.endGroup();
            acbSettings.beginGroup("Modified");
            acbSettings.setValue("LogoCoordsX", currentItemX_);
            acbSettings.setValue("LogoCoordsY", currentItemY_);
            acbSettings.endGroup();
        }

        // writeAction:
        Action writeAction(QStringLiteral("org.kde.altcusbranding.write"));
        writeAction.setHelperId(QStringLiteral("org.kde.altcusbranding"));
        QVariantMap args;
        args[QStringLiteral("imageFile")] = newImage;
        args[QStringLiteral("grubImageFilePng")] = resizedImage;
        args[QStringLiteral("grubImageFileJpg")] = convertedImage;
        writeAction.setArguments(args);
        writeAction.setParentWidget(this);
        writeAction.setTimeout(300000); //msec. Helper timeout min - 2 minutes

        ExecuteJob *mainJob = writeAction.execute(Action::AuthorizeOnlyMode);
        if (!mainJob->exec())
        {
            qInfo() << " *** Error code:       " << mainJob->error();
            qInfo() << " *** Error description:" << mainJob->errorText();
            QMessageBox::information(this, "Error",
                QString("KAuth returned an error code: %1 \n%2").arg(mainJob->error()).arg(mainJob->errorString()));
            int strTest = Action::AuthorizedStatus;
            qInfo() << " *** Authorized status: " << strTest;
        }
        else
        {
            ui->label_wait->setText(QString(tr("Please wait!")));
            mainJob = writeAction.execute();
            connect(mainJob, SIGNAL(result(KJob*)), this, SLOT(jobFinished(KJob*)));
            mainJob->start();
        }
    }
}

QString MainWindow::convertImageToJpg(QString newImage)
{
    // convert -colorspace YCbCr  -sampling-factor 2x2 newImage JPEG:grubImageFile
    QString grubImageFile = appTmpSettingsDir_ + QStringLiteral("newimage800x600.jpg");
    QFile grubImgFile(grubImageFile);
    if (grubImgFile.exists())
    {
        grubImgFile.remove();
    }
    QString convertComand = QStringLiteral("convert -colorspace YCbCr  -sampling-factor 2x2 ") + newImage + QStringLiteral(" JPEG:") + grubImageFile;
    int convertResult = system(qPrintable(convertComand));
    if (convertResult == 0)
    {
        qInfo() << " * Grub image file (jpeg) saved successfully.";
    }
    else
    {
        qInfo() << " * Grub image (jpeg) save error!";
    }
    return grubImageFile;
}

QString MainWindow::resizeImage(QString newImage, double grubImgWidth, double grubImgHeight)
{
    // Resise Image to 800x600. It needs to save grub image in "/boot/grub/themes/":
    QImage imgToResize(newImage);
    QImage imageFile800x600;
    double imgWidth = imgToResize.width();
    double imgHeight = imgToResize.height();
    double ratioImgToRes = imgWidth/imgHeight;
    double ratioStandart = grubImgWidth/grubImgHeight;
    if ( round(ratioImgToRes*1000)/1000 == round(ratioStandart*1000)/1000)
    {
        imageFile800x600 = imgToResize.scaled(grubImgWidth, grubImgHeight,
                                              Qt::KeepAspectRatio,
                                              Qt::SmoothTransformation);
    }
    else
    {
        QImage imageFileX = imgToResize.scaled(grubImgWidth, grubImgHeight,
                                               Qt::KeepAspectRatioByExpanding,
                                               Qt::SmoothTransformation);
        imageFile800x600 = imageFileX.copy( 0, 0, grubImgWidth, grubImgHeight);
    }
    QString grubImageFile = appTmpSettingsDir_ + QStringLiteral("newimage800x600.png");
    QFile grubImgFile(grubImageFile);
    if (grubImgFile.exists()) { grubImgFile.remove(); }

    bool res = imageFile800x600.save(grubImageFile, 0, -1);
    if (res)
    {
        qInfo() << " * Grub image file (png) saved successfully: " << grubImageFile;
    }
    else
    {
        qInfo() << " * Grub image (png) save error!";
    }
    return grubImageFile;
}


QString MainWindow::createNewImage(QString imageFile, QString logoFile) // Read all data from form and rename image files
{
    QImage imageIn(imageFile);
    QImage logoIn(logoFile);
    QString tempDir = appTmpSettingsDir_;
    bool res;
    // Add logo resize (left mouse button)
    QImage resultImage = mergeImages(imageIn, logoIn); // Combine two images in one
    QString newFileName = tempDir + QStringLiteral("newimage.png"); // Path to new file
    QFile someFile(newFileName);
    if (someFile.exists()) { someFile.remove(); }
    res = resultImage.save(newFileName, 0, -1);
    if (res)
    {
        qInfo() << " *** Image created success! ";
        return (newFileName);
    }
    else
    {
        QMessageBox::information(this, tr("Information"), tr("Image not created!"));
        qInfo() << " *** Image not created! ";
        return("false");
    }
}

QImage MainWindow::mergeImages(const QImage& backImage, const QImage& logoImage) // Combine two images (Logo + Background image)
{
    QImage resultImage = QImage(backImage.size(), QImage::Format_ARGB32_Premultiplied);
    QPainter painter(&resultImage);
    painter.setCompositionMode(QPainter::CompositionMode_Source);
    painter.fillRect(resultImage.rect(), Qt::transparent);
    painter.setCompositionMode(QPainter::CompositionMode_SourceOver); // background
    painter.drawImage(0, 0, backImage);
    painter.setCompositionMode(QPainter::CompositionMode_SourceOver); // logo image
    painter.drawImage(currentItemX_, currentItemY_, logoImage);
    painter.end();
    return resultImage;
}

void MainWindow::viewPlymouthImageF() // Add Plymouth image to scene
{
    viewFileToChangeBranding(ui->lineEdit_plymouth->text());
    if ((!logoInScene_)&&(addLogoButtonActivated_ == true)) // If logo not in Scene
    {
        viewLogoToChangeBranding();
    }
}

void MainWindow::viewNewLogoF() // Add new Logo to scene
{
    if (logoInScene_) // If logo in Scene, clean old logo, text and repaint background
    {
        killTimer(myTimerId_);
        logoInScene_ = false;
        addLogoButtonActivated_ = false;
        if (setScene_)
        {
            ui->graphicsView->scene()->clear();
            QString imageFile = ui->lineEdit_plymouth->text();
            if (!imageFile.isEmpty() && addBackgroundBtnActivated_)
            {
                viewFileToChangeBranding(ui->lineEdit_plymouth->text());
            }
        }
    }
    viewLogoToChangeBranding();
    myTimerId_ = startTimer(30); // Start timer.
    QString logoPath = ui->lineEdit_newLogo->text();
    if (!logoPath.isEmpty())
    {
        addLogoButtonActivated_ = true; // Logo in scene.
    }
    else addLogoButtonActivated_ = false;
}

void MainWindow::openPlymouthImageF() // Choose new Plymouth image
{
    QString fileTCBPath = openFileToChangeBranding(ui->lineEdit_plymouth->text());
    ui->lineEdit_plymouth->setText(fileTCBPath);
}

void MainWindow::openNewLogoF() // Choose new logo
{
    QString fileTCBPath = openFileToChangeBranding(ui->lineEdit_newLogo->text());
    ui->lineEdit_newLogo->setText(fileTCBPath);
}

void MainWindow::closeApplication() // Cancel button
{
    killTimer(myTimerId_);
    this->close();
}

void MainWindow::cleanupForm() // Clear data from all element on/in user form
{
    ui->lineEdit_plymouth->clear();
    ui->lineEdit_newLogo->clear();
    if (setScene_)
    {
        ui->graphicsView->scene()->clear();
    }
    else
    {
        qInfo() << "No images to clean.";
    }
    killTimer(myTimerId_);
    fileTCBPathInScene_ = "";
    logoInScene_ = false;
    addLogoButtonActivated_ = false;
    addBackgroundBtnActivated_ = false;
}

void MainWindow::timerEvent(QTimerEvent *event) // Timer Event
{
    if (event->timerId() == myTimerId_)
    {
        logoCoordsDetermination();
    }
    else
    {
        QWidget::timerEvent(event);
    }
}

void MainWindow::logoCoordsDetermination()
{
    if (logoInScene_)
    {
        currentItemX_ = item->x();
        currentItemY_ = item->y();
    }
}

void MainWindow::useTranslator(QTranslator* translator, const QString &domain)
{
    if (translator)
    {
        QCoreApplication::removeTranslator(translator);
        delete translator;
    }
    translator = new QTranslator(this);

    if (translator->load(QLibraryInfo::location(QLibraryInfo::TranslationsPath) + QStringLiteral("/") + domain + QStringLiteral("_") + QLocale::system().name()))
    {
        QCoreApplication::installTranslator(translator);
    }
    else
    {
        delete translator;
        translator = 0;
    }
}

void MainWindow::jobFinished(KJob* kjob)
{
    ExecuteJob *job2 = (ExecuteJob *)kjob;
    qInfo() << " *** Job error: " << job2->error();
    qInfo() << " *** Out data:  " << job2->data().value(QStringLiteral("output")).toByteArray();

    if (job2->error() != 0)
    {
        qInfo() << " *** Process failed!";
        QMessageBox::information(this, tr("Information"), tr("Process failed!"));
        qInfo() << " *** Error text:   " << job2->errorText();
        qInfo() << " *** Error string: " << job2->errorString();
        ui->label_wait->setText(QString(tr("Process failed")));
    }
    else
    {
        qInfo() << " *** Process completed successfully!";
        QMessageBox::information(this, tr("Information"), tr("Process completed successfully!"));
        ui->label_wait->setText(QString(tr("Process completed successfully")));
    }
}


